# Annotation

##################################################
# Prodigal

rule annotation_prodigal:
    input:
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.FILTERED.fasta")
    output:
        os.path.join(RESULTS_DIR, "annotation/prodigal/{rtype}/{tool}/proteins.faa")
    log:
        "logs/prodigal.{rtype}.{tool}.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads: 1
    conda:
        os.path.join(ENV_DIR, "prodigal.yaml")
    message:
        "Annotation: call genes w/ Prodigal"
    shell:
        "(date && prodigal -a {output} -p meta -i {input} && date) &> {log}"

##################################################
# DIAMOND

# NOTE: currently NOT used
# For info on "--longs-reads" option see https://github.com/bbuchfink/diamond/releases/tag/v0.9.23
rule annotation_diamond_lr_daa:
    input: 
        reads=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.fastq.gz"),
        db=config["diamond"]["db"]
    output:
        # NOTE: use uncompressed DAA format; other formats can be derived using `diamond view`
        os.path.join(RESULTS_DIR, "annotation/diamond/long_reads/proteins.daa")
    log:
        "logs/diamond_daa.long_reads.log",
    threads:
        config["diamond"]["threads"]
    conda:
        os.path.join(ENV_DIR, "diamond.yaml")
    message:
        "Annotation: protein search w/ DIAMOND in long reads"
    shell: 
        "(date && diamond blastx -q {input.reads} --db {input.db} --out {output} -p {threads} --long-reads --sensitive --outfmt 100 && date) &> {log}"

# TODO: This could maybe be parametrized by the respectively used DB
rule annotation_diamond_daa:
    input: 
        faa=os.path.join(RESULTS_DIR, "annotation/prodigal/{rtype}/{tool}/proteins.faa"),
        db=config["diamond"]["db"]
    output:
        # NOTE: use uncompressed DAA format; other formats can be derived using `diamond view`
        os.path.join(RESULTS_DIR, "annotation/diamond/{rtype}/{tool}/proteins.daa")
    log:
        "logs/diamond_daa.{rtype}.{tool}.log",
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads:
        config["diamond"]["threads"]
    conda:
        os.path.join(ENV_DIR, "diamond.yaml")
    message:
        "Annotation: protein search w/ DIAMOND"
    shell: 
        "(date && diamond blastp -q {input.faa} --db {input.db} --out {output} -p {threads} --outfmt 100 && date) &> {log}"

rule annotation_diamond_tsv:
    input:
        os.path.join(RESULTS_DIR, "annotation/diamond/{rtype}/{tool}/proteins.daa")
    output:
        os.path.join(RESULTS_DIR, "annotation/diamond/{rtype}/{tool}/proteins.tsv")
    log:
        "logs/diamond_tsv.{rtype}.{tool}.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads: 5
    params:
        # NOTE: For the used conda version, need to specify the entire column layout (outfmt) explicitly
        outfmt="6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen",
        ibname=lambda wildcards, input: os.path.splitext(input[0])[0]
    conda:
        os.path.join(ENV_DIR, "diamond.yaml")
    message:
        "Annotation: reformat DIAMOND output"
    shell:
        "(date && "
        "diamond view --daa {params.ibname} --max-target-seqs 1 -p {threads} --outfmt {params.outfmt} --out {output} && "
        "date) &> {log}"

##################################################
# RGI

rule annotation_rgi_input:
    input:
        os.path.join(RESULTS_DIR, "annotation/prodigal/{rtype}/{tool}/proteins.faa")
    output:
        temp(os.path.join(RESULTS_DIR, "annotation/rgi/{rtype}/{tool}/input.faa"))
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    message:
        "Annotation: RGI input"
    shell:
        # NOTE: remove stop codon symbol "*"
        "sed 's/\*$//' {input} > {output}"

# Run RGI: proteins
rule annotation_rgi:
    input:
        faa=os.path.join(RESULTS_DIR, "annotation/rgi/{rtype}/{tool}/input.faa"),
        db=os.path.join(DB_DIR, "rgi/card.json"),
        # NOTE: to make sure that the same DB is used for all targets
        setup="status/rgi_setup.done"
    output:
        os.path.join(RESULTS_DIR, "annotation/rgi/{rtype}/{tool}/rgi.txt")
    log:
        "logs/rgi.{rtype}.{tool}.log",
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads:
        config["rgi"]["threads"]
    params:
        alignment_tool=config["rgi"]["alignment_tool"],
        obname=lambda wildcards, output: os.path.splitext(output[0])[0]
    conda:
        os.path.join(ENV_DIR, "rgi.yaml")
    message:
        "Annotation: AMR prediction w/ RGI"
    shell:
        "(date && "
        "rgi database --version --local && "
        # NOTE: https://github.com/arpcard/rgi/issues/93: KeyError: 'snp'
        #       need to run the CMD twice
        "rgi main --input_sequence {input.faa} --output_file {params.obname} --input_type protein --local -a {params.alignment_tool} --clean -n {threads} || "
        "rgi main --input_sequence {input.faa} --output_file {params.obname} --input_type protein --local -a {params.alignment_tool} --clean -n {threads} && "
        "date) &> {log}"

##################################################
# CRISPR

rule annotation_casc:
    input:
        asm=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.FILTERED.fasta"),
        bin=os.path.join(MOD_DIR, "casc/bin/casc")
    output:
        os.path.join(RESULTS_DIR, "annotation/casc/{rtype}/{tool}/ASSEMBLY.FILTERED.results.txt")
    log:
        "logs/casc.{rtype}.{tool}.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads:
        config["casc"]["threads"]
    conda:
        os.path.join(ENV_DIR, "casc.yaml")
    message:
        "Annotation: CRISPR detection w/ CASC"
    shell:
        "(date && "
        "export PATH=$PATH:$(dirname {input.bin}) && "
        "export PERL5LIB=$(dirname $(dirname {input.bin}))/lib && "
        "casc -i {input.asm} -o $(dirname {output}) -n {threads} --conservative && "
        "date) &> {log}"

rule annotation_minced:
    input:
        asm=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.FILTERED.fasta"),
        jar=os.path.join(MOD_DIR, "minced/minced.jar")
    output:
        txt=os.path.join(RESULTS_DIR, "annotation/minced/{rtype}/{tool}/ASSEMBLY.FILTERED.results.txt"),
        gff=os.path.join(RESULTS_DIR, "annotation/minced/{rtype}/{tool}/ASSEMBLY.FILTERED.results.gff")
    log:
        "logs/minced.{rtype}.{tool}.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    conda:
        os.path.join(ENV_DIR, "minced.yaml")
    message:
        "Annotation: CRISPR detection w/ MinCED"
    shell:
        "(date && "
        "export PATH=$PATH:$(dirname {input.jar}) && "
        "minced {input.asm} {output.txt} {output.gff} && "
        "date) &> {log}"

##################################################
# Plasmids

rule annotation_plasflow_input:
    input:
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.FILTERED.fasta")
    output:
        temp(os.path.join(RESULTS_DIR, "annotation/plasflow/{rtype}/{tool}/input.fasta"))
    log:
        "logs/plasflow_input.{rtype}.{tool}.log",
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads: 1
    params:
        script=os.path.join(SRC_DIR, "filter_fasta_by_length.pl"),
        minlen=config["plasflow"]["minlen"]
    message:
        "Annotation: PlasFlow input"
    shell:
        "{params.script} {params.minlen} {input} > {output} 2> {log}"

rule annotation_plasflow:
    input:
        os.path.join(RESULTS_DIR, "annotation/plasflow/{rtype}/{tool}/input.fasta")
    output:
        tmp=temp(os.path.join(RESULTS_DIR, "annotation/plasflow/{rtype}/{tool}/plasflow.tsv.tmp")),
        tsv=os.path.join(RESULTS_DIR, "annotation/plasflow/{rtype}/{tool}/plasflow.tsv")
    log:
        "logs/plasflow.{rtype}.{tool}.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS)
    threads: 1
    params:
        threshold=config["plasflow"]["threshold"]
    conda:
        os.path.join(ENV_DIR, "plasflow.yaml")
    message:
        "Annotation: plasmid prediction w/ PlasFlow"
    shell:
        "(date && "
        "PlasFlow.py --input {input} --output {output.tmp} --threshold {params.threshold} && "
        "cut -f3,4,6- {output.tmp} > {output.tsv} && "
        "date) &> {log}"

##################################################
# rRNA genes

rule annotation_barrnap:
    input:
        os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.FILTERED.fasta")
    output:
        fa=os.path.join(RESULTS_DIR, "annotation/barrnap/{rtype}/{tool}/{kingdom}.fa"),
        gff=os.path.join(RESULTS_DIR, "annotation/barrnap/{rtype}/{tool}/{kingdom}.gff")
    log:
        "logs/barrnap.{rtype}.{tool}.{kingdom}.log"
    wildcard_constraints:
        rtype="|".join(READ_TYPES),
        tool="|".join(ASSEMBLERS),
        kingdom="|".join(config["barrnap"]["kingdom"])
    threads:
        config["barrnap"]["threads"]
    conda:
        os.path.join(ENV_DIR, "barrnap.yaml")
    message:
        "Annotation: rRNA genes w/ barrnap"
    shell:
        "(date && barrnap --threads {threads} --kingdom {wildcards.kingdom} --outseq {output.fa} < {input} > {output.gff} && date) &> {log}"